



#define pin_led 13

#define pin_event1 2

#define pin_event2 3



int batteries;
int serialactive;

#include <stdarg.h>

void pr(char *fmt, ... )
{
  if(!serialactive) return;
  
  char buf[128]; // resulting string limited to 128 chars
  va_list args;
  va_start (args, fmt );
  vsnprintf(buf, 128, fmt, args);
  va_end (args);
  Serial.println(buf);
}




// https://code.google.com/p/tinkerit/wiki/SecretVoltmeter

long readVcc() {
  long result;
  // Read 1.1V reference against AVcc
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA,ADSC));
  result = ADCL;
  result |= ADCH<<8;
  result = 1126400L / result; // Back-calculate AVcc in mV
  return result;
}






int runningoffbatteries(void)
{
 if(readVcc()<4200) return(1);
 return(0);  
}









#include <SerialCommand.h>

SerialCommand sCmd;

#include <Time.h>


time_t basetime;

int interruptevent;

#define INT_TIME 0x1
#define INT_PULSE_1 0x2
#define INT_PULSE_2 0x4


#include <EEPROM.h>

#define EEPROMSIZE 1024


#define LOGNULL 0x0
#define LOGMARK 0x1
#define LOGTIME 0x2
#define LOGOK  0x3
#define LOGFAIL  0x4


static const char * logtable[]={"null","mark","time","OK","Fail"};

#define LOGMAX (sizeof(logtable)/sizeof(char *))

// log entries start with LOGMARK and go backwards to LOGNULL or LOGMARK

static int logposn;

long eread(int i)
{
 long n; 
 n=EEPROM.read(i);
 i=(i+1) & (EEPROMSIZE-1);
 n|=((long)EEPROM.read(i))<<8;
 i=(i+1) & (EEPROMSIZE-1);  
 n|=((long)EEPROM.read(i))<<16;  
 i=(i+1) & (EEPROMSIZE-1);  
 n|=((long)EEPROM.read(i))<<24;
 
 return(n);
}


void ewrite(int i,long val)
{
 EEPROM.write(i,val & 0xFF);  
 i=(i+1) & (EEPROMSIZE-1);
 EEPROM.write(i,(val>>8) & 0xFF); 
 i=(i+1) & (EEPROMSIZE-1);
 EEPROM.write(i,(val>>16) & 0xFF); 
 i=(i+1) & (EEPROMSIZE-1);
 EEPROM.write(i,(val>>24) & 0xFF);  
}



void clearlog(void)
{
 int i;
 long w;

 randomSeed(micros()); 

 for(i=0;i<EEPROMSIZE;i++)
 { 
  w=eread(i);
  if(w==LOGMARK) ewrite(i,LOGNULL);
 }
 
 i=random(EEPROMSIZE);
 ewrite(i,LOGNULL);
 i=(i+sizeof(long)) & (EEPROMSIZE-1);
 ewrite(i,LOGMARK);
 logposn=i;
}

void showlog(void)
{
 int  i;
 byte b; 
 int  n;

 char * arg; 
 int    steps;
 
 arg = sCmd.next();
 if(arg!=NULL)
 {
  steps=atoi(arg);
 }
 else
 {
  steps=0x7FFF; 
 }

  
 pr("Show log");
  
 i=logposn;
 n=0;
 
 while(1)
 {
  i=(i-1);
  i= i & (EEPROMSIZE-1);
  b=EEPROM.read(i); //pr("log byte %d",b);
  if(b==LOGMARK || b==LOGNULL || i==logposn) break;
  if(b>=LOGMAX) break;
  if(b==LOGTIME)
  {
   i=(i-sizeof(long));
   i= i & (EEPROMSIZE-1); 
   time_t t=eread(i);
   n++;
   tmElements_t tm;
   breakTime(t,tm);  
   pr("%02d-%02d-%02d %02d/%02d/%04d",tm.Hour,tm.Minute,tm.Second,tm.Day,tm.Month,tm.Year+1970);
  }
  else
  {  
   pr("%02d %s",n,logtable[b]);
  } 
  if(n>=steps) break;
 } 
}


void eventlog(int event)
{
 if(event==LOGTIME)
 {
  ewrite(logposn,now());pr("wr x %lx",now());
  logposn=(logposn+sizeof(long)) & (EEPROMSIZE-1);   
 }
 EEPROM.write(logposn,event);
 logposn=(logposn+1) & (EEPROMSIZE-1);
 ewrite(logposn,LOGMARK);
}


void initlog(void)
{
 int  i;
 long w;
 
 for(i=0;i<EEPROMSIZE;i++)
 { 
  w=eread(i);
  if(w==LOGMARK)
  {
   break; 
  }  
 }

 if(i>=EEPROMSIZE) clearlog();  
 else              logposn=i;
}





void cmdwriteseconds(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  time_t t=now();
  setTime(hour(t),minute(t),atoi(arg),day(t),month(t),year(t));
  basetime=now();
 }
}



void cmdwriteminutes(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  time_t t=now();
  setTime(hour(t),atoi(arg),second(t),day(t),month(t),year(t));
  basetime=now();
 }
}




void cmdwritehours(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  time_t t=now();
  setTime(atoi(arg),minute(t),second(t),day(t),month(t),year(t));
  basetime=now();  
 }
}

void cmdwritedate(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  time_t t=now();
  setTime(hour(t),minute(t),second(t),atoi(arg),month(t),year(t));
  basetime=now();  
 }
}



void cmdwritemonth(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  time_t t=now();
  setTime(hour(t),minute(t),second(t),day(t),atoi(arg),year(t));
  basetime=now();  
 }
}


void cmdwriteyear(void)
{
 char * arg;
 arg = sCmd.next();
 if(arg!=NULL)
 {
  time_t t=now();
  setTime(hour(t),minute(t),second(t),day(t),month(t),atoi(arg));
  basetime=now();  
 }
}


void cmdreadtime(void)
{
 tmElements_t tm;

 breakTime(now(),tm);  
  
 pr("%02d-%02d-%02d %02d/%02d/%04d",tm.Hour,tm.Minute,tm.Second,tm.Day,tm.Month,tm.Year+1970);
  
}

void cmdreadbasetime(void)
{
  
 pr("%lu",basetime); 
  
}


void cmdunrecognized(const char *command)
{
 pr("What?"); 
}  


void readpower(void)
{
 pr("v=%lu batteries %d",readVcc(),batteries);
}

void clocksetup(void)
{
 sCmd.addCommand("rb",cmdreadbasetime);
 sCmd.addCommand("rt",cmdreadtime);
 sCmd.addCommand("ws",cmdwriteseconds);
 sCmd.addCommand("wm",cmdwriteminutes);
 sCmd.addCommand("wh",cmdwritehours);
 sCmd.addCommand("wd",cmdwritedate); 
 sCmd.addCommand("wn",cmdwritemonth); 
 sCmd.addCommand("wy",cmdwriteyear); 
 sCmd.addCommand("sl",showlog); 
 sCmd.addCommand("cl",clearlog); 
 sCmd.addCommand("rp",readpower);   
  
 sCmd.setDefaultHandler(cmdunrecognized); 
 
 setTime(0,0,0,22,7,2015);
 basetime=now();
  
}    








#include <avr/sleep.h>
#include <avr/power.h>
#include <avr/interrupt.h>







void ext_int_1(void) 
{
 
 interruptevent|=INT_PULSE_1;
}


void ext_int_2(void) 
{
 
 interruptevent|=INT_PULSE_2;
}



void power_down(void)
{
 if(serialactive)
 {
  serialactive=0;
  Serial.flush();
  Serial.end();
 } 
 pinMode(1, INPUT);
}

void power_up(void)
{
 serialactive=1; 
 Serial.begin(9600);        // connect to the serial port  
 pr("start up");
}






#ifndef sleep_bod_disable() // not included in Arduino AVR toolset
#define sleep_bod_disable() \
do { \
  uint8_t tempreg; \
  __asm__ __volatile__("in %[tempreg], %[mcucr]" "\n\t" \
                       "ori %[tempreg], %[bods_bodse]" "\n\t" \
                       "out %[mcucr], %[tempreg]" "\n\t" \
                       "andi %[tempreg], %[not_bodse]" "\n\t" \
                       "out %[mcucr], %[tempreg]" \
                       : [tempreg] "=&d" (tempreg) \
                       : [mcucr] "I" _SFR_IO_ADDR(MCUCR), \
                         [bods_bodse] "i" (_BV(BODS) | _BV(BODSE)), \
                         [not_bodse] "i" (~_BV(BODSE))); \
} while (0)
#endif



#define OVERFLOWMODE 1

//#define CTCMODE 1

#ifdef CTCMODE
ISR(TIMER2_COMPA_vect)
{
 basetime+=8; 
 interruptevent|=INT_TIME;  
}
#endif

#ifdef OVERFLOWMODE
ISR(TIMER2_OVF_vect)
{
 basetime+=8;
 interruptevent|=INT_TIME;  
}
#endif


byte adcsra;

void power_down_lowlevel(void)
{
 adcsra = ADCSRA;                  //save the ADC Control and Status Register A
 ADCSRA = 0;
 power_adc_disable ();
 power_spi_disable();
 power_twi_disable();
 power_timer0_disable();
 power_timer1_disable();
 power_usart0_disable();
}





void power_up_lowlevel(void)
{
 power_timer0_enable();
 power_timer1_enable();
 power_usart0_enable();
 power_adc_enable();  
 ADCSRA = adcsra;
 power_spi_enable();
 power_twi_enable();  
}  





  
  
  
  
void setup() 
{
 batteries=runningoffbatteries();
 

#ifdef CTCMODE  
  // clock input to timer 2 from XTAL1/XTAL2
 ASSR = bit (AS2);  

  // set up timer 2 to count up to 32 * 1024  (32768)
 TCCR2A = bit (WGM21);                             // CTC
 TCCR2B = bit (CS20) | bit (CS21) | bit (CS22);    // Prescaler of 1024                                  
 OCR2A =  255;              // count to 32 (zero-relative)                  

  // enable timer interrupts
 TIMSK2 |= bit (OCIE2A);
  
#endif

#ifdef OVERFLOWMODE

 //Set up TIMER2
    TIMSK2 = 0;                        //stop timer2 interrupts while we set up
    ASSR = _BV(AS2);                   //Timer/Counter2 clocked from external crystal
    TCCR2A = 0;                        //override arduino settings, ensure WGM mode 0 (normal mode)
    TCCR2B = _BV(CS22) | _BV(CS21) | _BV(CS20);    //prescaler clk/1024 -- TCNT2 will overflow once every 8 seconds
    TCNT2 = 0;                         //start the timer at zero
    while (ASSR & (_BV(TCN2UB) | _BV(TCR2AUB) | _BV(TCR2BUB))) {}    //wait for the registers to be updated    
    TIFR2 = _BV(OCF2B) | _BV(OCF2A) | _BV(TOV2);                     //clear the interrupt flags
    TIMSK2 = _BV(TOIE2);               //enable interrupt on overflow

#endif
  
  

 clocksetup();
 initlog();
 attachInterrupt(0, ext_int_1, CHANGE);
 attachInterrupt(1, ext_int_2, CHANGE);
 
 pinMode(pin_event1, INPUT);
 pinMode(pin_event2, INPUT);
 
 if(batteries) power_down(); 
 else          power_up();
}


void gotosleep(void)
{
 power_down_lowlevel();
  
 cli();
 set_sleep_mode(SLEEP_MODE_PWR_SAVE);  
 sleep_enable();
 sleep_bod_disable();
 sei();
 sleep_cpu();

// time passes

 sleep_disable();

 power_up_lowlevel();
 
 setTime(basetime); 
}







 
void loop() 
{
 int newbatteries=runningoffbatteries(); 

 
 if(newbatteries!=batteries)
 {
  if(newbatteries) power_down(); 
  else             power_up();
  batteries=newbatteries;
 }
 
  
 if(batteries)
 {

  gotosleep(); 
 }
 else
 { 
  if(Serial.available()>0) 
  {
   sCmd.readSerial();
  }

 }
 
 if(interruptevent & INT_TIME)
 {
  interruptevent&=~INT_TIME;

  pr("tick");
  setTime(basetime);


 }
 
 
 if(interruptevent & INT_PULSE_1)
 {
   
  pr("interrupt event");
   
  interruptevent&=~INT_PULSE_1;
 }


 if(interruptevent & INT_PULSE_2)
 {
  
  pr("on off pulse");
   
  interruptevent&=~INT_PULSE_2; 
 }

}
  
 
 
  
